/* FWServ - IRC Services
** bconf.c - Configuration implementation for the csircd derived bind
**           style conf
**
** Copyright J. Johnston and others.  See README for more details
** Some code Copyright: Jonathan George, Kai Seidler, ircd-hybrid Team,
**                      IRCnet IRCD developers.
**
** $Id: bconf.c,v 1.4 2003/11/21 18:32:22 ircd Exp $
*/

#include "configparse.h"
#include "fwserv.h"
#include "token.h"
#include "flags.h"
#include "nickserv.h"
#include "chanserv.h"

void AddOperHost (struct Operator *, char *);

#define PARAM_ERROR(ce, s)	{ log(1, "%s:%i: no parameter for %s " \
                                    "config option: %s", \
                                    (ce)->ce_fileptr->cf_filename, \
                                    (ce)->ce_varlinenum, (s), (ce)->ce_varname); \
  return 1; }

struct ConfTable
{
  char *name;
  int rehashable;
  int (*handler)(CONFIGENTRY *);
};

void AddOperHost (struct Operator *op, char *host)
{
    dlink_node *dl;
    dl = dlink_create ();
    dlink_add (strdup(host), dl, &op->hosts);
}

/* XXX Add support for (s|m|hr|d) */

static int generic_subblock_handler(CONFIGENTRY *, char *, struct ConfTable *);

static int c_unused	(CONFIGENTRY *);
static int c_serverinfo	(CONFIGENTRY *);
static int c_mysql (CONFIGENTRY *);

#ifdef NICKSERVICES
static int c_nickserv	(CONFIGENTRY *);
#ifdef CHANNELSERVICES
static int c_chanserv	(CONFIGENTRY *);
#endif
#ifdef MEMOSERVICES
static int c_memoserv	(CONFIGENTRY *);
#endif
#ifdef BOTSERVICES
static int c_botserv	(CONFIGENTRY *);
#endif
#endif

static int c_operserv	(CONFIGENTRY *);
static int c_operator	(CONFIGENTRY *);

static struct ConfTable conf_root_table[] =
{
  { "SERVERINFO",	1,	c_serverinfo },
  { "MYSQL",		1,	c_mysql },

#ifdef NICKSERVICES
  { "NICKSERV",		1,	c_nickserv },
#ifdef CHANNELSERVICES
  { "CHANSERV",		1,	c_chanserv },
#endif
#ifdef MEMOSERVICES
  { "MEMOSERV",		1,	c_memoserv },
#endif
#ifdef BOTSERVICES
  { "BOTSERV",		1,	c_botserv },
#endif
#endif

  { "OPERSERV",		1,	c_operserv },
  { "OPERATOR",		1,	c_operator },

  { NULL, 0, NULL }
};

static int c_si_name	    (CONFIGENTRY *);
static int c_si_desc	    (CONFIGENTRY *);
static int c_si_silent	    (CONFIGENTRY *);
static int c_si_uplink	    (CONFIGENTRY *);
static int c_si_pass	    (CONFIGENTRY *);
static int c_si_port	    (CONFIGENTRY *);
static int c_si_reconnect   (CONFIGENTRY *);
static int c_si_hidden	    (CONFIGENTRY *);
static int c_si_adminname   (CONFIGENTRY *);
static int c_si_adminemail  (CONFIGENTRY *);
static int c_si_privmsg	    (CONFIGENTRY *);
static int c_si_loglevel	(CONFIGENTRY *);

static struct ConfTable conf_serverinfo_table[] =
{
  { "NAME",	    0,	c_si_name },
  { "DESC",  1,	c_si_desc },
  { "SILENT",	    1,	c_si_silent },

  { "UPLINK",	    1,  c_si_uplink },
  { "PASS",	    1,	c_si_pass },
  { "PORT",	    1,	c_si_port },

  { "RECONNECT",    1,	c_si_reconnect },
  { "HIDDEN",	    1,	c_si_hidden },

  { "ADMINNAME",    1,	c_si_adminname },
  { "ADMINEMAIL",   1,	c_si_adminemail },

  { "PRIVMSG",	    1,	c_si_privmsg },
  { "LOGLEVEL",		1,	c_si_loglevel },

  { NULL, 0, NULL }
};

static int c_sql_host (CONFIGENTRY *);
static int c_sql_user (CONFIGENTRY *);
static int c_sql_pass (CONFIGENTRY *);
static int c_sql_db (CONFIGENTRY *);

static struct ConfTable conf_mysql_table[] =
{
  { "HOST",	1,	c_sql_host },
  { "USER",	1,	c_sql_user },
  { "PASS",	1,	c_sql_pass },
  { "DB",	1,	c_sql_db },

  { NULL,	0,	NULL }
};

#ifdef NICKSERVICES
static int c_nickserv_nick  (CONFIGENTRY *);
static int c_nickserv_user  (CONFIGENTRY *);
static int c_nickserv_host  (CONFIGENTRY *);
static int c_nickserv_name  (CONFIGENTRY *);
static int c_nickserv_flags (CONFIGENTRY *);
static int c_nickserv_expire (CONFIGENTRY *);
#ifdef USE_MYSQL
static int c_nickserv_table (CONFIGENTRY *);
#endif

static struct ConfTable conf_nickserv_table[] =
{
  { "NICK",	0,	c_nickserv_nick },
  { "USER",	0,	c_nickserv_user	},
  { "HOST",	0,	c_nickserv_host },
  { "NAME",	0,	c_nickserv_name },

  { "FLAGS",	1,	c_nickserv_flags },
  { "EXPIRE",	1,	c_nickserv_expire },

#ifdef USE_MYSQL
  { "TABLE",	1,	c_nickserv_table },
#endif

  { NULL,	0,	NULL }
};

#ifdef CHANNELSERVICES
static int c_chanserv_nick  (CONFIGENTRY *);
static int c_chanserv_user  (CONFIGENTRY *);
static int c_chanserv_host  (CONFIGENTRY *);
static int c_chanserv_name  (CONFIGENTRY *);
#ifdef USE_MYSQL
static int c_chanserv_table (CONFIGENTRY *);
static int c_chanserv_access_table (CONFIGENTRY *);
static int c_chanserv_levels_table (CONFIGENTRY *);
#endif

static struct ConfTable conf_chanserv_table[] =
{
  { "NICK",     0,      c_chanserv_nick },
  { "USER",     0,      c_chanserv_user },
  { "HOST",     0,      c_chanserv_host },
  { "NAME",     0,      c_chanserv_name },

#ifdef USE_MYSQL
  { "TABLE",	1,	c_chanserv_table },
  { "ACCESS_TABLE",	1,	c_chanserv_access_table },
  { "LEVELS_TABLE", 1,	c_chanserv_levels_table },
#endif

  { NULL,       0,      NULL }
};
#endif /* CHANNELSERVICES */

#ifdef MEMOSERVICES
static int c_memoserv_nick  (CONFIGENTRY *);
static int c_memoserv_user  (CONFIGENTRY *);
static int c_memoserv_host  (CONFIGENTRY *);
static int c_memoserv_name  (CONFIGENTRY *);
#ifdef USE_MYSQL
static int c_memoserv_table (CONFIGENTRY *);
#endif

static struct ConfTable conf_memoserv_table[] =
{
  { "NICK",     0,      c_memoserv_nick },
  { "USER",     0,      c_memoserv_user },
  { "HOST",     0,      c_memoserv_host },
  { "NAME",     0,      c_memoserv_name },

#ifdef USE_MYSQL
  { "TABLE",	1,	c_memoserv_table },
#endif
  { NULL,       0,      NULL }
};
#endif /* MEMOSERVICES */

#ifdef BOTSERVICES
static int c_botserv_nick  (CONFIGENTRY *);
static int c_botserv_user  (CONFIGENTRY *);
static int c_botserv_host  (CONFIGENTRY *);
static int c_botserv_name  (CONFIGENTRY *);
#ifdef USE_MYSQL
static int c_botserv_table  (CONFIGENTRY *);
static int c_botserv_channel_table (CONFIGENTRY *);
#endif

static struct ConfTable conf_botserv_table[] =
{
  { "NICK",     0,      c_botserv_nick },
  { "USER",     0,      c_botserv_user },
  { "HOST",     0,      c_botserv_host },
  { "NAME",     0,      c_botserv_name },

#ifdef USE_MYSQL
  { "TABLE",	1,	c_botserv_table },
  { "CHANNEL_TABLE",	1,	c_botserv_channel_table },
#endif

  { NULL,       0,      NULL }
};
#endif /* BOTSERVICES */
#endif /* NICKSERVICES */

static int c_operserv_nick  (CONFIGENTRY *);
static int c_operserv_user  (CONFIGENTRY *);
static int c_operserv_host  (CONFIGENTRY *);
static int c_operserv_name  (CONFIGENTRY *);
static int c_operserv_channels	(CONFIGENTRY *);
static int c_operserv_operonly	(CONFIGENTRY *);

static struct ConfTable conf_operserv_table[] =
{
  { "NICK",     0,      c_operserv_nick },
  { "USER",     0,      c_operserv_user },
  { "HOST",     0,      c_operserv_host },
  { "NAME",     0,      c_operserv_name },
  { "CHANNELS",	0,	c_operserv_channels },
  { "OPERONLY",	1,	c_operserv_operonly },

  { NULL,       0,      NULL }
};

/* Set default parameters.  The settings used are the same as
** what is provided in stats.cfg.dist except for the binary
** ones, which MUST be default 0 in order for the conf load
** to work properly
*/
void InitDefaults()
{
    /* These strcpy's are safe, as they are static strings */
    strcpy (me.name, "irc.services");
    strcpy (me.desc, "IRC Services");

    strcpy (me.uplink, "irc.hub");
    strcpy (me.pass, "uplink");

    strcpy (me.adminname, "Misconfigured Administrator");
    strcpy (me.adminemail, "misconfigured@nodomain.com");

    me.silent = 0;
    me.port = 6667;
    me.reconnect = 30;
    me.hidden = 0;
    me.privmsg = 0;
	me.loglevel = L_INFO;

#ifdef NICKSERVICES
    strcpy (NickServ.nick, "NickServ");
    strcpy (NickServ.user, "NickServ");
    strcpy (NickServ.host, me.name);
    strcpy (NickServ.name, "Nickname Services");

    NickServ.expire = 2592000;
    NickServ.flags |= (NS_SECURE|NS_ALLOWMEMO|NS_MEMOSIGNON|NS_MEMONOTIFY|NS_AUTOMASK);

#ifdef USE_MYSQL
    strcpy (NickServ.table, "NickServ");
#endif

#ifdef CHANNELSERVICES
    strcpy (ChanServ.nick, "ChanServ");
    strcpy (ChanServ.user, "ChanServ");
    strcpy (ChanServ.host, me.name);
    strcpy (ChanServ.name, "Channel Services");

    ChanServ.expire = 2592000;
    ChanServ.flags |= (CS_GUARD);
#ifdef USE_MYSQL
    strcpy (ChanServ.table, "ChanServ");
	strcpy (ChanServ.access_table, "ChannelAccess");
	strcpy (ChanServ.levels_table, "ChannelLevels");
#endif
#endif

#ifdef MEMOSERVICES
    strcpy (MemoServ.nick, "MemoServ");
    strcpy (MemoServ.user, "MemoServ");
    strcpy (MemoServ.host, me.name);
    strcpy (MemoServ.name, "Memo Services");

    MemoServ.expire = 2419200;
#ifdef USE_MYSQL
    strcpy (MemoServ.table, "MemoServ");
#endif
#endif

#ifdef BOTSERVICES
    strcpy (BotServ.nick, "BotServ");
    strcpy (BotServ.user, "BotServ");
    strcpy (BotServ.host, me.name);
    strcpy (BotServ.name, "Bot Services");
#ifdef USE_MYSQL
    strcpy (BotServ.table, "BotServ");
	strcpy (BotServ.channel_table, "BotChannels");
#endif
#endif
#endif /* NICKSERVICES */

    strcpy (OperServ.nick, "OperServ");
    strcpy (OperServ.user, "OperServ");
    strcpy (OperServ.host, me.name);
    strcpy (OperServ.name, "Operator Services");

    OperServ.operonly = 0;
    OperServ.save = 1800;

}

/* Used to verify setting from the conf during startup and rehash...
** This prevents invalid variables from being used
*/
void VerifyConf()
{

	if ((me.loglevel  == TOKEN_UNMATCHED) || (me.loglevel == TOKEN_ERROR))
		me.loglevel = L_INFO;

}

void ConfLoad(void)
{
  CONFIGFILE *cfptr;
  CONFIGFILE *cfp;
  CONFIGENTRY *ce;
  struct ConfTable *ct = NULL;

  cfptr = cfp = config_load(CPATH);
  if (cfp == NULL)
  {
    fprintf(stderr, "Unable to open configuration file (%s): %s\n",
            CPATH, strerror(errno));
    /*close_logfiles();*/
    exit(0);
  }
  for (;cfptr;cfptr = cfptr->cf_next) /* Handle include files? */
  {
    for (ce = cfptr->cf_entries;ce;ce = ce->ce_next)
    {
      for (ct = conf_root_table; ct->name; ct++)
      {
        if (!strcasecmp(ct->name, ce->ce_varname))
        {
          ct->handler(ce);
          break;
        }
      }
    }
    if (ct->name == NULL)
    {
      log(1, "%s:%i: invalid config option: %s",
           ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
           ce->ce_varname);
    }
  }
  config_free(cfp);
}

void rehash(void)
{
  CONFIGFILE *cfptr;
  CONFIGFILE *cfp;
  CONFIGENTRY *ce;
  struct ConfTable *ct = NULL;

  cfptr = cfp = config_load(CPATH);
  if (cfp == NULL)
  {
    log (1, "Unable to open configuration file (%s): %s\n", CPATH, strerror(errno));
  }
  for (;cfptr;cfptr = cfptr->cf_next) /* Handle include files? */
  {
    for (ce = cfptr->cf_entries;ce;ce = ce->ce_next)
    {
      for (ct = conf_root_table; ct->name; ct++)
      {
        if (!strcasecmp(ct->name, ce->ce_varname))
        {
          if (ct->rehashable == 1)
            ct->handler(ce);
          break;
        }
      }
    }
    if (ct->name == NULL)
    {
      log(1, "%s:%i: invalid config option: %s",
           ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
           ce->ce_varname);
    }
  }
  config_free(cfp);

  VerifyConf();

  /* And finally (just in case the times changed) reset the
  ** conf time based events
  */
  /*ResetTimedEvents();*/
}

static int generic_subblock_handler(CONFIGENTRY *ce, char *subblock,
                                    struct ConfTable *table)
{
  struct ConfTable *ct = NULL;

  for (ce = ce->ce_entries;ce;ce = ce->ce_next)
  {
    for (ct = table; ct->name; ct++)
    {
      if (!strcasecmp(ct->name, ce->ce_varname))
      {
        ct->handler(ce);
        break;
      }
    }
    if (ct->name == NULL)
    {
      log(1, "%s:%i: invalid %s config option: %s",
           ce->ce_fileptr->cf_filename, ce->ce_varlinenum,
           subblock, ce->ce_varname);
    }
  }
  return 0;
}

static int c_serverinfo (CONFIGENTRY *ce)
{
	generic_subblock_handler(ce, "SERVERINFO", conf_serverinfo_table);
	return 0;
}

static int c_mysql (CONFIGENTRY *ce)
{
	generic_subblock_handler (ce, "MYSQL", conf_mysql_table);
	return 0;
}

#ifdef NICKSERVICES
static int c_nickserv (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "NICKSERV", conf_nickserv_table);
    return 0;
}

#ifdef CHANNELSERVICES
static int c_chanserv (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "CHANSERV", conf_chanserv_table);
    return 0;
}
#endif

#ifdef MEMOSERVICES
static int c_memoserv (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "MEMOSERV", conf_memoserv_table);
    return 0;
}
#endif

#ifdef BOTSERVICES
static int c_botserv (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "BOTSERV", conf_botserv_table);
    return 0;
}
#endif
#endif /* NICKSERVICES */

static int c_operserv (CONFIGENTRY *ce)
{
    generic_subblock_handler (ce, "OPERSERV", conf_operserv_table);
    return 0;
}

/* Local server settings */
static int c_si_name (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    strlcpy (me.name, ce->ce_vardata, sizeof(me.name));
    return 0;
}

static int c_si_desc (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    strlcpy (me.desc, ce->ce_vardata, sizeof(me.desc));
    return 0;
}

static int c_si_reconnect (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    me.reconnect = ce->ce_vardatanum;
    return 0;
}

static int c_si_silent (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    me.silent = ce->ce_vardatanum;
    return 0;
}

static int c_si_adminname (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    strlcpy (me.adminname, ce->ce_vardata, sizeof(me.adminname));
    return 0;
}

static int c_si_adminemail (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    strlcpy (me.adminemail, ce->ce_vardata, sizeof(me.adminemail));
    return 0;
}

static int c_si_privmsg (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "serverinfo");
    }
    me.privmsg = ce->ce_vardatanum;
    return 0;
}

static int c_si_loglevel (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "serverinfo");
	}
	me.loglevel = token_to_value (log_flags, ce->ce_vardata);
	return 0;
}

/* Uplink Information */
static int c_si_uplink (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    strlcpy (me.uplink, ce->ce_vardata, sizeof(me.uplink));
    return 0;
}

static int c_si_port (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    me.port = ce->ce_vardatanum;
    return 0;
}

static int c_si_pass (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    strlcpy (me.pass, ce->ce_vardata, sizeof(me.pass));
    return 0;
}

static int c_si_hidden (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR(ce, "serverinfo");
    }
    me.hidden = ce->ce_vardatanum;
    return 0;
}

static int c_sql_host (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR(ce, "mysql");
	}
	strlcpy (me.mhost, ce->ce_vardata, sizeof(me.mhost));
	return 0;
}

static int c_sql_user (CONFIGENTRY *ce)
{
        if (ce->ce_vardata == NULL)
        {
                PARAM_ERROR(ce, "mysql");
        }
        strlcpy (me.muser, ce->ce_vardata, sizeof(me.muser));
        return 0;
}

static int c_sql_pass (CONFIGENTRY *ce)
{
        if (ce->ce_vardata == NULL)
        {
                PARAM_ERROR(ce, "mysql");
        }
        strlcpy (me.mpass, ce->ce_vardata, sizeof(me.mpass));
        return 0;
}

static int c_sql_db (CONFIGENTRY *ce)
{
        if (ce->ce_vardata == NULL)
        {
                PARAM_ERROR(ce, "mysql");
        }
        strlcpy (me.mdb, ce->ce_vardata, sizeof(me.mdb));
        return 0;
}

/*
static int c_listen (CONFIGENTRY *ce)
{
	char host[63];
	int ipcount;

	for (ce = ce->ce_entries;ce;ce = ce->ce_next)
	{
		if (!strcasecmp(ce->ce_varname, "HOST"))
		{
			ipcount = 0;
			strlcpy (host, ce->ce_vardata, sizeof(host));
			AddListener (host);
		}
		else if (!strcasecmp(ce->ce_varname, "PORT"))
		{
			AddPortToHost (host, ce->ce_vardata);
			ipcount++;
		}
	}

	return 0;
}
*/

#ifdef NICKSERVICES
/* NickServ Settings */
static int c_nickserv_nick (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "nickserv");
    }
    strlcpy (NickServ.nick, ce->ce_vardata, sizeof(NickServ.nick));
    return 0;
}

static int c_nickserv_user (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "nickserv");
    }
    strlcpy (NickServ.user, ce->ce_vardata, sizeof(NickServ.user));
    return 0;
}

static int c_nickserv_host (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "nickserv");
    }
    strlcpy (NickServ.host, ce->ce_vardata, sizeof(NickServ.host));
    return 0;
}

static int c_nickserv_name (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "nickserv");
    }
    strlcpy (NickServ.name, ce->ce_vardata, sizeof(NickServ.name));
    return 0;
}

static int c_nickserv_flags (CONFIGENTRY *ce)
{
    if (ce->ce_entries == NULL)
    {
	log (1, "%s:%i: no entries for nickserv config option: %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
	return 1;
    }

    for (ce = ce->ce_entries; ce; ce = ce->ce_next)
    {
	int val;
	val = token_to_value (nick_flags, ce->ce_varname);
	if ((val != TOKEN_UNMATCHED) && (val != TOKEN_ERROR))
	    NickServ.flags |= val;
	else
	    log (1, "%s:%i: operator flags: Unknown flag: %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
    }
    return 0;
}

static int c_nickserv_expire (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "nickserv");
    }
    NickServ.expire = ce->ce_vardatanum;
    return 0;
}

#ifdef USE_MYSQL
static int c_nickserv_table (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "nickserv");
    }
    strlcpy (NickServ.table, ce->ce_vardata, sizeof(NickServ.table));
    return 0;
}
#endif

#ifdef CHANNELSERVICES
/* ChanServ Settings */
static int c_chanserv_nick (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "chanserv");
    }
    strlcpy (ChanServ.nick, ce->ce_vardata, sizeof(ChanServ.nick));
    return 0;
}

static int c_chanserv_user (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "chanserv");
    }
    strlcpy (ChanServ.user, ce->ce_vardata, sizeof(ChanServ.user));
    return 0;
}

static int c_chanserv_host (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "chanserv");
    }
    strlcpy (ChanServ.host, ce->ce_vardata, sizeof(ChanServ.host));
    return 0;
}

static int c_chanserv_name (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "chanserv");
    }
    strlcpy (ChanServ.name, ce->ce_vardata, sizeof(ChanServ.name));
    return 0;
}

#ifdef USE_MYSQL
static int c_chanserv_table (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "chanserv");
    }
    strlcpy (ChanServ.table, ce->ce_vardata, sizeof(ChanServ.table));
    return 0;
}

static int c_chanserv_access_table (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "chanserv");
	}
	strlcpy (ChanServ.access_table, ce->ce_vardata, sizeof(ChanServ.access_table));
	return 0;
}

static int c_chanserv_levels_table (CONFIGENTRY *ce)
{   
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "chanserv");
    }
    strlcpy (ChanServ.levels_table, ce->ce_vardata, sizeof(ChanServ.levels_table));
    return 0;
}
#endif

#endif /* CHANNELSERVICES */

#ifdef MEMOSERVICES
/* MemoServ Settings */
static int c_memoserv_nick (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "memoserv");
    }
    strlcpy (MemoServ.nick, ce->ce_vardata, sizeof(MemoServ.nick));
    return 0;
}

static int c_memoserv_user (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "memoserv");
    }
    strlcpy (MemoServ.user, ce->ce_vardata, sizeof(MemoServ.user));
    return 0;
}

static int c_memoserv_host (CONFIGENTRY *ce)
{   
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "memoserv");
    }
    strlcpy (MemoServ.host, ce->ce_vardata, sizeof(MemoServ.host));
    return 0;
}

static int c_memoserv_name (CONFIGENTRY *ce)
{   
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "memoserv");
    }
    strlcpy (MemoServ.name, ce->ce_vardata, sizeof(MemoServ.name));
    return 0;
}

#ifdef USE_MYSQL
static int c_memoserv_table (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "memoserv");
    }
    strlcpy (MemoServ.table, ce->ce_vardata, sizeof(MemoServ.table));
    return 0;
}
#endif
#endif /* MEMOSERVICES */

#ifdef BOTSERVICES
/* BotServ Settings */
static int c_botserv_nick (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "botserv");
    }
    strlcpy (BotServ.nick, ce->ce_vardata, sizeof(BotServ.nick));
    return 0;
}

static int c_botserv_user (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "botserv");
    }
    strlcpy (BotServ.user, ce->ce_vardata, sizeof(BotServ.user));
    return 0;
}

static int c_botserv_host (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "botserv");
    }
    strlcpy (BotServ.host, ce->ce_vardata, sizeof(BotServ.host));
    return 0;
}

static int c_botserv_name (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "botserv");
    }
    strlcpy (BotServ.name, ce->ce_vardata, sizeof(BotServ.name));
    return 0;
}

#ifdef USE_MYSQL
static int c_botserv_table (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "botserv");
    }
    strlcpy (BotServ.table, ce->ce_vardata, sizeof(BotServ.table));
    return 0;
}

static int c_botserv_channel_table (CONFIGENTRY *ce)
{
	if (ce->ce_vardata == NULL)
	{
		PARAM_ERROR (ce, "botserv");
	}
	strlcpy (BotServ.channel_table, ce->ce_vardata, sizeof(BotServ.channel_table));
	return 0;
}
#endif
#endif /* BOTSERVICES */
#endif /* NICKSERVICES */

/* OperServ Settings */
static int c_operserv_nick (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "operserv");
    }
    strlcpy (OperServ.nick, ce->ce_vardata, sizeof(OperServ.nick));
    return 0;
}

static int c_operserv_user (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "operserv");
    }
    strlcpy (OperServ.user, ce->ce_vardata, sizeof(OperServ.user));
    return 0;
}

static int c_operserv_host (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "operserv");
    }
    strlcpy (OperServ.host, ce->ce_vardata, sizeof(OperServ.host));
    return 0;
}

static int c_operserv_name (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
        PARAM_ERROR (ce, "operserv");
    }
    strlcpy (OperServ.name, ce->ce_vardata, sizeof(OperServ.name));
    return 0;
}

static int c_operserv_channels (CONFIGENTRY *ce)
{
    if (ce->ce_entries == NULL)
    {
	log (1, "%s:%i: no entries for operserv config option: %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
	return 1;
    }

    for (ce = ce->ce_entries; ce; ce = ce->ce_next)
    {
	AddOperServChannel (ce->ce_varname);
    }
    return 0;
}

static int c_operserv_operonly (CONFIGENTRY *ce)
{
    if (ce->ce_vardata == NULL)
    {
	PARAM_ERROR (ce, "operserv");
    }
    OperServ.operonly = ce->ce_vardatanum;
    return 0;
}

/* Operator Blocks */
static int c_operator (CONFIGENTRY *ce)
{
    struct Operator *op;
    CONFIGENTRY *flce;
    dlink_node *dl;

    op = calloc(1, sizeof(struct Operator));

    if (ce->ce_vardata != NULL)
	strlcpy (op->nick, ce->ce_vardata, sizeof(op->nick));

    for (ce = ce->ce_entries; ce; ce = ce->ce_next)
    {
	if (!strcasecmp(ce->ce_varname, "NAME"))
	{
	    if (ce->ce_vardata == NULL)
	    {
		log (1, ":%s:%i: no parameter for operator config option: %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
		free (op);
		return 1;
	    }
	    strlcpy (op->nick, ce->ce_vardata, sizeof(op->nick));
	    continue;
	}
	else if(!strcasecmp(ce->ce_varname, "PASSWORD"))
	{
	    if (ce->ce_vardata == NULL)
	    {
		log (1, ":%s:%i: no parameter for operator config option: %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
		free (op);
		return 1;
	    }
	    strlcpy (op->password, ce->ce_vardata, sizeof(op->password));
	    continue;
	}
	else if(!strcasecmp(ce->ce_varname, "HOST"))
	{
	    if (ce->ce_vardata == NULL)
		continue;
	    AddOperHost (op, ce->ce_vardata);
	    continue;
	}
	else if(!strcasecmp(ce->ce_varname, "FLAGS"))
	{
	    for (flce = ce->ce_entries; flce; flce = flce->ce_next)
	    {
		int val;
		val = token_to_value (oper_flags, flce->ce_varname);
		if ((val != TOKEN_UNMATCHED) && (val != TOKEN_ERROR))
		    op->flags |= val;
		else
		    log (1, "%s:%i: operator flags: Unknown flag: %s", flce->ce_fileptr->cf_filename, flce->ce_varlinenum, flce->ce_varname);
	    }
	    continue;
	}
	else
	{
	    log (1, "%s:%i: invalid operator config option: %s", ce->ce_fileptr->cf_filename, ce->ce_varlinenum, ce->ce_varname);
	    continue;
	}
    }

    if (op->nick[0] == '\0')
    {
	log (1, "Incomplete operator block");
	free (op);
	return 1;
    }

    Mem.operators++;
    dl = dlink_create();
    dlink_add (op, dl, &me.operators);
    return 0;
}

static int c_unused (CONFIGENTRY *ce)
{
  return 0;
}
