#include "fwserv.h"
#include "config.h"
#include "utils.h"
#include "channel.h"
#include "user.h"
#include "data.h"
#include "chanserv.h"

#if defined(NICKSERVICES) && defined(CHANNELSERVICES)

void chanserv (struct User *, char *, char *);

void RegisterChannel (struct User *, struct Channel *, char *, char *);

static int cs_help (struct User *, int, char **);
static int cs_register (struct User *, int, char **);
static int cs_join (struct User *, int, char **);
static int cs_drop (struct User *, int, char **);
static int cs_setpass (struct User *, int, char **);
static int cs_part (struct User *, int, char **);
static int cs_access (struct User *, int, char **);
static int cs_set (struct User *, int, char **);
static int cs_op (struct User *, int, char **);
static int cs_level (struct User *, int, char **);

static int cs_set_guard (struct User *, int, char **);
static int cs_set_secure (struct User *, int, char **);
static int cs_set_topic		(struct User *, int, char **);
static int cs_set_topiclock	(struct User *, int, char **);

static struct CmdTable chanserv_table[] =
{
  { "HELP",	NULL,	cs_help,	LVL_NONE },

  { "ACCESS",	NULL,	cs_access,	LVL_IDENT },
  { "DROP",	NULL,	cs_drop,	LVL_IDENT },
  { "REGISTER",	NULL,	cs_register,	LVL_IDENT },

  { "JOIN",	NULL,	cs_join,	LVL_ADMIN },
  { "PART",	NULL,	cs_part,	LVL_ADMIN },
  { "SETPASS",	NULL,	cs_setpass,	LVL_ADMIN },

  { "SET",	NULL,	cs_set,		LVL_IDENT },

  { "OP",	NULL,	cs_op,		LVL_IDENT },
  { "LEVEL",	NULL,	cs_level,	LVL_IDENT },

  { (char *)0,  NULL,   (int (*)()) 0,	(int) 0}
};

static struct CmdTable set_table[] =
{
  { "GUARD",	NULL,	cs_set_guard,	LVL_IDENT },
  { "SECURE",	NULL,	cs_set_secure,	LVL_IDENT },
  { "TOPIC",	NULL,	cs_set_topic,	LVL_IDENT },
  { "TOPICLOCK",	NULL,	cs_set_topiclock,	LVL_IDENT },

  { (char *)0,	NULL,	(int (*)()) 0,	(int) 0 }
};

static struct HelpTable help_table[] =
{
  { "HELP",     "help/chanserv/help",   	LVL_NONE },

  { "ACCESS",	"help/chanserv/access",		LVL_IDENT },
  { "DROP",	"help/chanserv/drop",		LVL_IDENT },
  { "REGISTER",	"help/chanserv/register",	LVL_IDENT },

  { "JOIN",	"help/chanserv/join",		LVL_ADMIN },
  { "PART",	"help/chanserv/part",		LVL_ADMIN },
  { "SETPASS",	"help/chanserv/setpass",	LVL_ADMIN },

  { "SET",	"help/chanserv/set",		LVL_IDENT },

  { (char *)0,  (char *)0,              	(int)0 }
};

static struct HelpTable set_help[] =
{
  { "GUARD",	"help/chanserv/guard",		LVL_IDENT },
  { "SECURE",	"help/chanserv/secure",		LVL_IDENT },
};

static int cs_set (struct User *u, int parc, char *parv[])
{
	int i;
	if (parc < 3)
	{
		privmsg(u->nick, ChanServ.nick, "Syntax: \002SET <channel> <option> [parameter]\002");
		return;
	}

	for (i=0; set_table[i].handler; i++)
	{
		if (!strcasecmp(parv[2], set_table[i].name))
		{
			if ((set_table[i].flag & LVL_IDENT) && !(IsIdent(u)))
			{
				privmsg(u->nick, ChanServ.nick, RPL_MUST_IDENTIFY, set_table[i].name);
				return 0;
			}
			set_table[i].handler(u, parc, parv);
			return;
		}
	}

	privmsg(u->nick, ChanServ.nick, "Unknown SET option [\002%s\002]", parv[2]);
	return 0;
}

static int cs_set_guard (struct User *u, int parc, char *parv[])
{
	struct RChannel *rchan;

	if ((rchan = FindRChannel(parv[1])) == NULL)
	{
		privmsg(u->nick, ChanServ.nick, "Channel \002%s\002 not registered.");
		return 0;
	}

	if (parc < 4)
	{
		if (IsGuardSet(rchan))
		{
			privmsg(u->nick, ChanServ.nick, "ChanGuard for channel %s is [\002On\002]", rchan->name);
			return 0;
		}
		else
		{
			privmsg(u->nick, ChanServ.nick, "ChanGuard for channel %s is [\002Off\002]", rchan->name);
			return 0;
		}
		return 0;
	}

	if (!strcasecmp("ON", parv[3]))
	{
		if (IsGuardSet(rchan))
		{
			privmsg(u->nick, ChanServ.nick, "ChanGuard is already enabled for channel %s", rchan->name);
			return 0;
		}
		else
		{
			sts(":%s SJOIN %ld %s +nt :%s", me.name, time(NULL), rchan->name, ChanServ.nick);
			sts(":%s MODE %s +o %s", me.name, rchan->name, ChanServ.nick);
			privmsg(u->nick, ChanServ.nick, "ChanGuard for channel %s is now [\002On\002]", rchan->name);
			rchan->flags |= CS_GUARD;
			return 0;
		}
	}
	else if (!strcasecmp("OFF", parv[3]))
	{
		if (!IsGuardSet(rchan))
		{
			privmsg(u->nick, ChanServ.nick, "ChanGuard is already disabled for channel %s", rchan->name);
			return 0;
		}
		else
		{
			sts(":%s PART %s :Leaving", ChanServ.nick, rchan->name);
			privmsg(u->nick, ChanServ.nick, "ChanGuard for channel %s in now [\002Off\002]", rchan->name);
			rchan->flags &= ~CS_GUARD;
			return 0;
		}
	}
	else
	{
		privmsg(u->nick, ChanServ.nick, "Invalid parameter for SET GUARD");
		return 0;
	}

  return 0;
}

static int cs_set_topiclock (struct User *u, int parc, char *parv[])
{
    struct RChannel *rchan;

    if ((rchan = FindRChannel(parv[1])) == NULL)
    {
        privmsg(u->nick, ChanServ.nick, "Channel \002%s\002 not registered.");
        return 0;
    }

    if (parc < 4)
    {
        if (IsTopicLock(rchan))
        {
            privmsg(u->nick, ChanServ.nick, "TopicLock for channel %s is [\002On\002]", rchan->name);
            return 0;
        }
        else
        {
            privmsg(u->nick, ChanServ.nick, "TopicLock for channel %s is [\002Off\002]", rchan->name);
            return 0;
        }
        return 0;
    }

    if (!strcasecmp("ON", parv[3]))
    {
        if (IsTopicLock(rchan))
        {
            privmsg(u->nick, ChanServ.nick, "TopicLock is already enabled for channel %s", rchan->name);
            return 0;
        }
        else
        {
            privmsg(u->nick, ChanServ.nick, "TopicLock for channel %s is now [\002On\002]", rchan->name);
            rchan->flags |= CS_TOPICLOCK;
            return 0;
        }
    }
    else if (!strcasecmp("OFF", parv[3]))
    {
        if (!IsTopicLock(rchan))
        {
            privmsg(u->nick, ChanServ.nick, "TopicLock is already disabled for channel %s", rchan->name);
            return 0;
        }
        else
        {
            privmsg(u->nick, ChanServ.nick, "TopicLock for channel %s in now [\002Off\002]", rchan->name);
            rchan->flags &= ~CS_TOPICLOCK;
            return 0;
        }
    }
    else
    {
        privmsg(u->nick, ChanServ.nick, "Invalid parameter for SET TOPICLOCK");
        return 0;
    }

  return 0;
}

static int
cs_set_secure (struct User *u, int parc, char *parv[])
{
    struct RChannel *rchan;

    if ((rchan = FindRChannel(parv[1])) == NULL)
    {
        privmsg(u->nick, ChanServ.nick, "Channel \002%s\002 not registered.");
        return 0;
    }

    if (parc < 4)
    {
        if (IsSecure(rchan))
        {
            privmsg(u->nick, ChanServ.nick, "Secure for channel %s is [\002On\002]", rchan->name);
            return 0;
        }
        else
        {
            privmsg(u->nick, ChanServ.nick, "Secure for channel %s is [\002Off\002]", rchan->name);
            return 0;
        }
        return 0;
    }
    
    if (!strcasecmp("ON", parv[3]))
    {   
        if (IsSecure(rchan))
        {   
            privmsg(u->nick, ChanServ.nick, "Secure is already enabled for channel %s", rchan->name);
            return 0;
        }
        else
        {   
            privmsg(u->nick, ChanServ.nick, "Secure for channel %s is now [\002On\002]", rchan->name);
            rchan->flags |= CS_SECURE;
            return 0;
        }
    }
    else if (!strcasecmp("OFF", parv[3]))
    {   
        if (!IsSecure(rchan))
        {   
            privmsg(u->nick, ChanServ.nick, "Secure is already disabled for channel %s", rchan->name);
            return 0;
        }
        else
        {   
            privmsg(u->nick, ChanServ.nick, "Secure for channel %s in now [\002Off\002]", rchan->name);
            rchan->flags &= ~CS_SECURE;
            return 0;
        }
    }
    else
    {   
        privmsg(u->nick, ChanServ.nick, "Invalid parameter for SET SECURE");
        return 0;
    }
  
  return 0;
}

static int cs_set_topic (struct User *u, int parc, char *parv[])
{
	struct RChannel *rchan;
	char	topic[TOPICLEN];

	if (parc < 4)
	{
		privmsg(u->nick, ChanServ.nick, "Syntax: SET <channel> TOPIC <topic>");
		return;
	}

	if ((rchan = FindRChannel(parv[1])) != NULL)
	{
		if (!strcasecmp(u->nick, rchan->founder->nick))
		{
			topic[0] = '\0';
			ExpandParv (topic, TOPICLEN, 3, parc, parv);
			strlcpy (rchan->topic, topic, TOPICLEN);
			SetTopic(rchan);
			return 0;
		}
		privmsg(u->nick, ChanServ.nick, "Permission Denied");
		return 0;
	}

	privmsg(u->nick, ChanServ.nick, "Channel \002%s\002 not registered.");
	return 0;
}

static int
cs_level (struct User *u, int parc, char *parv[])
{
	char channel[CHANLEN];
	struct RChannel *rchan;

	if (parc < 3)
	{
		privmsg(u->nick, ChanServ.nick, "Syntax: LEVEL <channel> [LIST | SET | RESET] <type> <level>");
		return 0;
	}

	strlcpy (channel, parv[1], sizeof(channel));

	rchan = FindRChannel (channel);

	if (!rchan)
    {
        privmsg(u->nick, ChanServ.nick, "Channel \002%s\002 isn't registered.", channel);
        return 0;
    }

	if (!strcasecmp("LIST", parv[2]))
	{
		privmsg(u->nick, ChanServ.nick, "Levels for %s", channel);
		privmsg(u->nick, ChanServ.nick, "AUTODEOP %d", rchan->levels->autodeop);
		privmsg(u->nick, ChanServ.nick, "AUTOVOICE %d", rchan->levels->autovoice);
		privmsg(u->nick, ChanServ.nick, "CMDVOICE %d", rchan->levels->cmdvoice);
		privmsg(u->nick, ChanServ.nick, "AUTOOP %d", rchan->levels->autoop);
		privmsg(u->nick, ChanServ.nick, "CMDOP %d", rchan->levels->cmdop);
		privmsg(u->nick, ChanServ.nick, "AUTOHOP %d", rchan->levels->autohop);
		privmsg(u->nick, ChanServ.nick, "CMDHOP %d", rchan->levels->cmdhop);
	}
	return 0;
}

static int
cs_access (struct User *u, int parc, char *parv[])
{
	struct RChannel *rchan;
	struct AccessEntry *aentry;
	char	channel[CHANLEN];
	dlink_node *dl;
	int	count = 0, access = 0;

	if (parc < 3)
	{
		privmsg(u->nick, ChanServ.nick, "Syntax: ACCESS <channel> [LIST | ADD | DEL] <nickname> <level>");
		return 0;
	}

	strlcpy(channel, parv[1], CHANLEN+1);

	rchan = FindRChannel(channel);

	if (!rchan)
	{
		privmsg(u->nick, ChanServ.nick, "Channel \002%s\002 isn't registered.", channel);
		return 0;
	}

	access = GetAccess (rchan, u->nick);

	if (!strcasecmp("LIST", parv[2]))
	{
		privmsg(u->nick, ChanServ.nick, RPL_ACCESSLISTSTART, channel);
		DLINK_FOREACH (dl, rchan->access.head)
		{
			count++;
			aentry = (struct AccessEntry *)dl->data;
			privmsg(u->nick, ChanServ.nick, RPL_ACCESSLIST, count, aentry->level, strlen(aentry->nick), aentry->nick, "");
		}
		privmsg(u->nick, ChanServ.nick, RPL_ACCESSLISTEND);
		return;
	}
	else if(!strcasecmp("ADD", parv[2]))
	{
		if (parc < 5)
		{
			privmsg(u->nick, ChanServ.nick, "Syntax: ACCESS <channel> ADD <nickname> <level>");
			return 0;
		}

		AddAccess (rchan, parv[3], atoi(parv[4]));
		privmsg(u->nick, ChanServ.nick, "Added \002%s\002 with access level %s", parv[3], parv[4]);
		return 0;
	}
	else if(!strcasecmp("DEL", parv[2]))
	{
		if (parc < 4)
		{
			privmsg(u->nick, ChanServ.nick,	"Syntax: ACCESS <channel> DEL <nickname>");
			return 0;
		}

		DelAccess (rchan, parv[3]);
		privmsg(u->nick, ChanServ.nick, "Removed \002%s\002", parv[3]);
		return 0;
	}
	else
	{
		privmsg(u->nick, ChanServ.nick, "Invalid option for ACCESS: %s", parv[2]);
		return 0;
	}

	return 0;
}

static int
cs_part (struct User *u, int parc, char *parv[])
{
  return;
}

static int
cs_drop (struct User *u, int parc, char *parv[])
{
  return;
}

static int
cs_setpass (struct User *u, int parc, char *parv[])
{
  return;
}

static int cs_op (struct User *u, int parc, char *parv[])
{
	struct RChannel *rchan;
	struct ChanUser *cuser;
	char	channel[CHANLEN];
	int		i;

	if (parc < 2)
	{
		privmsg(u->nick, ChanServ.nick, "Syntax: OP <ALL|channel> [nick]");
		return 0;
	}

	if (!strcasecmp("ALL", parv[1]))
		return;
	else
	{
		strlcpy (channel, parv[1], CHANLEN+1);

		if ((cuser = FindChanUser (FindChannel(channel), u->nick)) == NULL)
		{
			privmsg (u->nick, ChanServ.nick, "Not on channel %s", channel);
			return;
		}

		rchan = FindRChannel(channel);

		if (!rchan)
		{
			privmsg(u->nick, ChanServ.nick, "Channel \002%s\002 isn't registered.", channel);
			return 0;
		}

		if ((GetAccessLevel(rchan, u->nick) < 10) && !IsFounder(rchan, u))
		{
			privmsg(u->nick, ChanServ.nick, "Level 10 is required for OP");
			return 0;
		}

		if (parc < 3)
		{
			/*if (IsOp(FindChannel(channel), u->nick) < 1)*/
			if (IsChanOp(cuser))
			{
				privmsg(u->nick, ChanServ.nick, "Your already OPed in %s", channel);
				return 0;
			}
			sts(":%s MODE %s +o %s", ChanServ.nick, rchan->name, u->nick);
			return 0;
		}

		for (i=2; parv[i] != NULL; i++)
		{
			/*if (IsOp(FindChannel(channel), parv[i]))*/
			if (IsChanOp(FindChanUser(FindChannel(channel), parv[i])))
				continue;
			else
				sts(":%s MODE %s +o %s", ChanServ.nick, rchan->name, parv[i]);
		}
	}

	return 0;
}

void
RegisterChannel (struct User *u, struct Channel *c, char *pass, char *desc)
{
	struct RChannel *rchan;
	struct AccessEntry *aentry;
	dlink_node *dl, *adl;

#ifdef CRYPT_CHANPASS
    char * salt;
    salt = make_salt();
#endif

	rchan = calloc(1, sizeof(struct RChannel));

	strlcpy (rchan->name, c->name, sizeof(rchan->name));
#ifdef CRYPT_CHANPASS
	strlcpy (rchan->pass, crypt(pass, salt), sizeof(rchan->pass));
#else
	strlcpy (rchan->pass, pass, sizeof(rchan->pass));
#endif

	if (desc)
		strlcpy (rchan->desc, desc, sizeof(rchan->desc));

	rchan->modes |= CHMODE_TOPIC|CHMODE_NOMSG;
	rchan->flags |= ChanServ.flags;
	rchan->founder = u->n;

	aentry = calloc(1, sizeof(struct AccessEntry));
	strlcpy(aentry->nick, u->n->nick, sizeof(aentry->nick));
	aentry->level = 40;
	adl = dlink_create ();
	dlink_add (aentry, adl, &rchan->access);

	dl = dlink_create ();
	dlink_add (rchan, dl, &rchannel_list);

	if (IsGuardSet(rchan))
	{
		sts(":%s SJOIN %ld %s +nt :%s", me.name, c->ts, c->name, ChanServ.nick);
		sts(":%s MODE %s +o %s", me.name, c->name, ChanServ.nick);
	}

	log (L_INFO, "Channel [%s] registered to %s [%s!%s@%s]", c->name, u->nick, u->nick, u->username, u->hostname);
#ifdef SPY
	Spy (ChanServ.nick, "Channel [%s] registered to %s [%s!%s@%s]", c->name, u->nick, u->nick, u->username, u->hostname);
#endif

	privmsg (u->nick, ChanServ.nick, "Channel [\002%s\002] registered with the password [\002%s\002]", c->name, pass);
	return;

}


static int
cs_join (struct User *u, int parc, char *parv[])
{
  if (parc != 2)
  {
    sts(":%s NOTICE %s :Syntax: JOIN <#channel>", ChanServ.nick, u->nick);
    return;
  }

  sts(":%s SJOIN %ld %s + :@%s", me.name, time(NULL), parv[1], ChanServ.nick);
}

static int
cs_register (struct User *u, int parc, char *parv[])
{
	/*struct NewChannel *c;*/
	char desc[BUFSIZE];
	char	channel[CHANLEN+1];
	struct	RChannel *rchan;
	struct	Channel *c;
	struct	ChanUser *cuser;

	if (parc < 3)
	{
		privmsg(u->nick, ChanServ.nick, "Syntax: \002REGISTER\002 <channel> <password> [description]");
		return 0;
	}

	strlcpy (channel, parv[1], CHANLEN+1);

	if (!is_channel(channel))
	{
		privmsg(u->nick, ChanServ.nick, "Channel name [\002%s\002] is invalid.", channel);
		return 0;
	}

	if ((rchan = FindRChannel(channel)) != NULL)
	{
		privmsg(u->nick, ChanServ.nick, "Channel [\002%s\002] already registered.", channel);
		return 0;
	}

	if ((c = FindChannel(channel)) != NULL)
	{
		/*if (!OnChannel(c, u->nick))*/
		if ((cuser = FindChanUser(c, u->nick)) == NULL)
		/*if (!IsOnChan(c, u->nick))*/
		{
			privmsg(u->nick, ChanServ.nick, "Your not joined \002%s\002", channel);
			return 0;
		}

		/*if (!IsOp(c, u->nick))*/
		if (!IsChanOp(cuser))
		{
			privmsg(u->nick, ChanServ.nick, "Your not an operator on [\002%s\002].", channel);
			return 0;
		}

		RegisterChannel (u, c, parv[2], (parv[3])? parv[3] : NULL);
		return 0;
	}

	privmsg(u->nick, ChanServ.nick, "Channel [\002%s\002] doesn't exist.", channel);
	return 0;

}

static int
cs_help (struct User *u, int parc, char *parv[])
{
  int i;
  
  if (parc < 2)
  {
    send_help_file(u, ChanServ.nick, "help/chanserv/index", "index");
    return;
  }

  if (!strcasecmp("SET", parv[1]))
  {
    if (parc < 3)
    {
      send_help_file(u, ChanServ.nick, "help/chanserv/set", "set");
      return;
    }

    for (i=0; set_help[i].name; i++)
    {
      if (!strcasecmp(parv[2], set_help[i].name))
      {
        send_help_file(u, ChanServ.nick, set_help[i].filename, parv[2]);
        return;
      }
    }
    notice(u, ChanServ.nick, "No help for SET %s", parv[2]);
    return;
  } 
  
  for (i=0; help_table[i].name; i++)
  {
    if (!strcasecmp(parv[1], help_table[i].name))
    {
      send_help_file(u, ChanServ.nick, help_table[i].filename, parv[1]);
      return;
    }
  }
  notice(u, ChanServ.nick, "No help for %s", parv[1]);
  return;
}

void chanserv(struct User *u, char *target, char *param)
{
  char *parv[MAXPARA];
  int parc, i, status;
  static char buf[BUFSIZE];
  parc = 0;
  
  for (i = 0; i < MAXPARA; i++)
    parv[i] = NULL;

  strlcpy(buf, param, BUFSIZE);

  if (param != NULL)
    parc = msgtoken(param, parv);

  if (parc == 0)
    return;
    
  for (i = 0; chanserv_table[i].handler; i++)
  {
    if(!strcasecmp(parv[0], chanserv_table[i].name))
    {
      if ((chanserv_table[i].flag & LVL_IDENT) && !(IsIdent(u)))
      {
	privmsg(u->nick, ChanServ.nick, RPL_MUST_IDENTIFY, chanserv_table[i].name);
        return;
      }

      if ((chanserv_table[i].flag & LVL_ADMIN) && !(u->flags & LVL_ADMIN))
      {
        privmsg(u->nick, ChanServ.nick, "Invalid Command: %s", parv[0]);
        return;
      }

      status = chanserv_table[i].handler(u, parc, parv);
      return;
    }
  }
 
  privmsg (u->nick, ChanServ.nick, "Invalid Command: %s", parv[0]);
  return;
}

void
JoinChannels()
{
  struct RChannel *c;
  dlink_node *dl;

  DLINK_FOREACH(dl, rchannel_list.head)
  {
    c = (struct RChannel *)dl->data;
	if (IsGuardSet(c))
	{
		ServerJoin(ChanServ.nick, c->name);
		sts(":%s MODE %s +o %s", me.name, c->name, ChanServ.nick);
	}

	if ((c->topic != NULL) && (c->topic[0] != '\0'))
		SetTopic(c);
  }

  return;
}

struct RChannel *FindRChannel (char *channel)
{
	struct RChannel *rchan;
	dlink_node *dl;

	DLINK_FOREACH (dl, rchannel_list.head)
	{
		rchan = (struct RChannel *)dl->data;
		if (!strcasecmp(rchan->name, channel))
			return rchan;
	}

	return NULL;
}

int GetAccessLevel (struct RChannel *rchan, char *nick)
{
	dlink_node *dl;
	struct AccessEntry *aentry;

	if (rchan == NULL)
		return -1;

	DLINK_FOREACH (dl, rchan->access.head)
	{
		aentry = (struct AccessEntry *)dl->data;
		if (!strcasecmp(aentry->nick, nick))
			return aentry->level;
	}

	if (IsFounder(rchan, nick))
		return 40;

	return 0;
}

int GetAccess (struct RChannel *rchan, char *nick)
{
    dlink_node *dl;
    struct AccessEntry *aentry;

    DLINK_FOREACH (dl, rchan->access.head)
    {
        aentry = (struct AccessEntry *)dl->data;
        if (!strcasecmp(aentry->nick, nick))
            return aentry->level;
    }

    return 0;
}

int IsFounder (struct RChannel *rchan, char *nick)
{
	if (!strcasecmp(rchan->founder->nick, nick))
		return 1;

	return 0;
}

int AddAccess (struct RChannel *rchan, char *nick, int level)
{
	dlink_node *dl;
	struct AccessEntry *aentry;

	DLINK_FOREACH (dl, rchan->access.head)
	{
		aentry = (struct AccessEntry *)dl->data;
		if (!strcasecmp(nick, aentry->nick))
			DelAccess (rchan, nick);
	}

	aentry = calloc (1, sizeof(struct AccessEntry));
	dl = dlink_create ();

	if (nick)
		strlcpy (aentry->nick, nick, sizeof(aentry->nick));
	aentry->level = level;

	dlink_add (aentry, dl, &rchan->access);
	return 0;
}

int DelAccess (struct RChannel *rchan, char *nick)
{
	dlink_node *dl, *tdl;
	struct AccessEntry *aentry;

	DLINK_FOREACH_SAFE (dl, tdl, rchan->access.head)
	{
		aentry = (struct AccessEntry *)dl->data;
		if (!strcasecmp(nick, aentry->nick))
		{
			dlink_delete (dl, &rchan->access);
			dlink_free (dl);
			free (aentry);
			return 0;
		}
	}

	return 0;
}

#endif /* defined(NICKSERVICES) && defined(CHANNELSERVICES) */
