#include "fwserv.h"
#include "config.h"
#include "nickserv.h"
#include "utils.h"
#include "data.h"
#include "user.h"

#ifdef USE_MYSQL
  #include <mysql.h>
  #include "_mysql.h"
#endif

#ifdef NICKSERVICES
void nickserv (struct User *, char *, char *);

static int ns_help (struct User *, int, char **);
static int ns_register (struct User *, int, char **);
static int ns_identify (struct User *, int, char **);
static int ns_drop (struct User *, int, char **);
static int ns_info (struct User *, int, char **);
static int ns_link (struct User *, int, char **);

static int ns_set (struct User *, int, char **);

static int ns_set_password (struct User *, int, char **);
static int ns_set_noexpire (struct User *, int, char **);
static int ns_set_email (struct User *, int, char **);
static int ns_set_url (struct User *, int, char **);
static int ns_set_secure (struct User *, int, char **);
static int ns_set_memos (struct User *, int, char **);
static int ns_set_signon (struct User *, int, char **);

static struct CmdTable nickserv_table[] =
{
 { "HELP",	NULL,	ns_help,	LVL_NONE },

 { "REGISTER",	NULL,	ns_register,	LVL_NONE },
 { "IDENTIFY",	NULL,	ns_identify,	LVL_NONE },
 { "DROP",	NULL,	ns_drop,	LVL_NONE },

 { "INFO",	NULL,	ns_info,	LVL_NONE },
 { "LINK",	NULL,	ns_link,	LVL_IDENT },

 { "SET",	NULL,	ns_set,		LVL_IDENT },
 { (char *)0,	NULL,	(int (*)()) 0,	(int) 0}
};

static struct CmdTable set_table[] =
{
  { "PASSWORD",	NULL,	ns_set_password,	LVL_IDENT },

  { "NOEXPIRE",	NULL,	ns_set_noexpire,	LVL_IDENT },
  { "EMAIL",	NULL,	ns_set_email,		LVL_IDENT },
  { "URL",	NULL,	ns_set_url,		LVL_IDENT },
  { "SECURE",	NULL,	ns_set_secure,		LVL_IDENT },
  { "MEMOS",	NULL,	ns_set_memos,		LVL_IDENT },
  { "SIGNON",	NULL,	ns_set_signon,		LVL_IDENT },
  { (char *)0,	NULL,	(int (*)()) 0,	(int) 0}
};

static struct HelpTable help_table[] =
{
  { "REGISTER",	"help/nickserv/register",	LVL_NONE },
  { "IDENTIFY", "help/nickserv/identify",	LVL_NONE },
  { "DROP",	"help/nickserv/drop",		LVL_NONE },
  { "INFO",	"help/nickserv/info",		LVL_NONE },
  { "LINK",	"help/nickserv/link",		LVL_NONE },
  { (char *)0,	(char *)0,			(int)0 }
};

static struct HelpTable set_help[] =
{
  { "PASSWORD",	"help/nickserv/set/password",	LVL_IDENT },
  { "NOEXPIRE",	"help/nickserv/set/noexpire",	LVL_ADMIN },
  { "EMAIL",	"help/nickserv/set/email",	LVL_IDENT },
  { "URL",	"help/nickserv/set/url",	LVL_IDENT },
  { "SECURE",	"help/nickserv/set/secure",	LVL_IDENT },
  { "MEMOS",	"help/nickserv/set/memos",	LVL_IDENT },
  { "SIGNON",	"help/nickserv/set/signon",	LVL_IDENT },
  { "KILL",	"help/nickserv/set/kill",	LVL_IDENT },

  { (char *)0,	(char *)0,			(int)0 }
};

void nickserv(struct User *u, char *target, char *param)
{
    char *parv[MAXPARA];
    int parc, i, status;
    static char buf[BUFSIZE];
    parc = 0;
  
    for (i = 0; i < MAXPARA; i++)
	parv[i] = NULL;
  
    strlcpy(buf, param, BUFSIZE);
  
    if (param != NULL)
	parc = msgtoken(param, parv);
  
    if (parc == 0)
	return;
  
    for (i = 0; nickserv_table[i].handler; i++)
    { 
	if(!strcasecmp(parv[0], nickserv_table[i].name))
	{
	    if ((nickserv_table[i].flag & LVL_IDENT) && !(IsIdent(u)))
	    {   
			privmsg(u->nick, NickServ.nick, RPL_MUST_IDENTIFY, nickserv_table[i].name);
			return;
	    }

	    if ((nickserv_table[i].flag & LVL_ADMIN) && !(IsOSAdmin(u)))
	    {
			privmsg(u->nick, NickServ.nick, "Invalid Command: %s", parv[0]);
			return;
	    }

	    status = nickserv_table[i].handler(u, parc, parv);
	    return;
	}
    }
  
    privmsg(u->nick, NickServ.nick, "Invalid Command: %s", parv[0]);
    return;
}

static int ns_link (struct User *u, int parc, char *parv[])
{
  struct LinkedNicks *link;
  struct Nick *n;
  dlink_node *dl;

  if (parc != 3)
  {
    return;
  }

  if ((n = FindNick(parv[1])) == NULL)
  {
    return;
  }

#ifdef CRYPT_NICKPASS
  if (!strcmp(n->password, crypt(parv[2], n->password)))
#else
  if (!strcmp(n->password, parv[2]))
#endif
  {
    /* Check for the link */
    if (IsLinked(n->nick, u->n->nick) > 0)
      /* Link exists */
      return;

    dl = dlink_create();
    link = calloc(1, sizeof(struct LinkedNicks));
    strcpy(link->master, n->nick);
    strcpy(link->nick, u->n->nick);

    dlink_add(link, dl, &linked_nicks);
  }

  notice(u, NickServ.nick, "%s Linked to %s", u->n->nick, n->nick);

  debug("D: %s %s %s", parv[0], parv[1], parv[2]);
  return;
}

static int
ns_help (struct User *u, int parc, char *parv[])
{
  int i;

  if (parc < 2)
  {
    send_help_file(u, NickServ.nick, "help/nickserv/index", "index");
    return;
  }

  if (!strcasecmp("SET", parv[1]))
  {
    if (parc < 3)
    {
      send_help_file(u, NickServ.nick, "help/nickserv/set/index", "SET index");
      return;
    }

    for (i=0; set_help[i].name; i++)
    {
      if (!strcasecmp(parv[2], set_help[i].name))
      {
        send_help_file(u, NickServ.nick, set_help[i].filename, parv[2]);
        return;
      }
    }

    notice(u, NickServ.nick, "No help for %s", parv[2]);
    return;
  }

  for (i=0; help_table[i].name; i++)
  {
    if (!strcasecmp(parv[1], help_table[i].name))
    {
      send_help_file(u, NickServ.nick, help_table[i].filename, parv[1]);
      return;
    }
  }
  notice(u, NickServ.nick, "No help for %s", parv[1]);
  return;
}

static int
ns_set (struct User *u, int parc, char *parv[])
{
  int i;
  if (parc < 2)
  {
    notice(u, NickServ.nick, "Syntax: \002SET <option> [parameter]\002");
    return;
  }

  for (i=0; set_table[i].handler; i++)
  {
    if (!strcasecmp(parv[1], set_table[i].name))
    {
      set_table[i].handler(u, parc, parv);
      return;
    }
  }

  notice(u, NickServ.nick, "Unknown switch [\002%s\002]", parv[1]);
  return 0;
}

static int
ns_set_password (struct User *u, int parc, char *parv[])
{
  struct Nick *n;
  char * salt;

  salt = make_salt();
  n = FindNick(u->nick);

  if (parc < 3)
  {
    notice(u, NickServ.nick, "Syntax: SET PASSWORD <password>");
    return;
  }

  if (n == NULL)
  {
    notice(u, NickServ.nick, "Nickname %s isn't registered", u->nick);
    return;
  }

  strcpy(n->password, crypt(parv[2], salt));
  notice(u, NickServ.nick, "Password set to %s, remember this for future", parv[2]);
}

static int
ns_set_noexpire (struct User *u, int parc, char *parv[])
{
  return;
}

static int
ns_set_email (struct User *u, int parc, char *parv[])
{
  struct Nick *n;

  if (parc < 3)
  {
    notice(u, NickServ.nick, "Syntax: SET EMAIL <email address>");
    return;
  }

  n = FindNick(u->nick);

  if (n == NULL)
  {
    notice(u, NickServ.nick, "Your nickname isn't registered");
    return;
  }
/*
  if (!Identified(nick))
  {
    notice(u, NickServ.nick, RPL_MUST_IDENTIFY, "SET");
    return;
  }
*/

  strcpy(n->email, parv[2]);
  notice(u, NickServ.nick, "Set email address to %s", parv[2]);
  return;
}

static int
ns_set_url (struct User *u, int parc, char *parv[])
{
  struct Nick *n;

  if (parc < 3)
  {
    notice(u, NickServ.nick, "Syntax: SET URL <url>");
    return;
  }

  n = FindNick(u->nick);

  if (n == NULL)
  {
    notice(u, NickServ.nick, "Your nickname isn't registered");
    return;
  }

/*
  if (!(Identified(nick)))
  {
    notice(u, n_NickServ, "Your not identified to NickServ");
    return;
  }
*/

  strcpy(n->url, parv[2]);
  notice(u, NickServ.nick, "Set URL to %s", parv[2]);
  return;
}

static int
ns_set_memos (struct User *u, int parc, char *parv[])
{
  return;
}

static int
ns_set_signon (struct User *u, int parc, char *parv[])
{
  struct Nick *n;

  if (parc < 2)
  {
    /* XXX - Change to a generic reply function */
    notice(u, NickServ.nick, "Syntax: SET SIGNON [ON|OFF]");
    return;
  }

  /* We assume that the nickname is registered
     and the user has identified to NickServ 
     Do we really need this? Because if the
     user is truely identified they have
     the NewNick struct in u */

  if ((n = FindNick(u->nick)) == NULL)
    return;

  /* User never implied ON or OFF so lets
     show them the status of it */

  if (parc < 3)
  {
    if (u->n->flags & NS_MEMOSIGNON)
    {
      notice(u, NickServ.nick, "Notification of Memos on Signon for your nickname is [\002ON\002]");
      return;
    }

    notice(u, NickServ.nick, "Notification of Memos on Signon for your nickname is [\002OFF\002]");
    return;
  }


  /* We have 2 options here either ON or OFF */

  if (!strcasecmp("ON", parv[2]))
  {
    if (!(u->n->flags & NS_MEMOSIGNON))
      u->n->flags |= NS_MEMOSIGNON;

    notice(u, NickServ.nick, "Toggled Notification of Memos on Signon [\002ON\002]");
    return;
  }
  else if (!strcasecmp("OFF", parv[2]))
  {
    if (u->n->flags & NS_MEMOSIGNON)
      u->n->flags &= ~NS_MEMOSIGNON;

    notice(u, NickServ.nick, "Toggled Notification of Memos on Signon [\002OFF\002]");
    return;
  } else {
    notice(u, NickServ.nick, "Syntax: \002SET SIGNON {ON|OFF}\002");
    return;
  }

  /* NOT REACHED */
  return 0;
}

static int
ns_set_secure (struct User *u, int parc, char *parv[])
{
  struct Nick *n;

  if (parc < 3)
  {
    notice(u, NickServ.nick, "Syntax: SET SECURE [ON|OFF]");
    return;
  }

  n = FindNick(u->nick);

  if ((!strcasecmp("ON", parv[2])) || (!strcasecmp("OFF", parv[2])))
  {
    if (!strcasecmp("ON", parv[2]))
    {
      if (!(n->flags & NS_SECURE))
      {
        n->flags |= NS_SECURE;
        notice(u, NickServ.nick, "SECURE [\002ON\002]");
        return;
      } else {
        notice(u, NickServ.nick, "SECURE already set");
        return;
      }
    } else if (!strcasecmp("OFF", parv[2])) {
        if (n->flags & NS_SECURE)
        {
          n->flags &= ~NS_SECURE;
          notice(u, NickServ.nick, "SECURE [\002OFF\002]");
          return;
        } else {
          notice(u, NickServ.nick, "SECURE not set.");
          return;
        }
    }
  } else {
    notice(u, NickServ.nick, "SECURE [ON|OFF]");
  }
}

static int
ns_register (struct User *u, int parc, char *parv[])
{
    struct Nick *n;

    if (parc != 2)
    {
	privmsg (u->nick, NickServ.nick, "Syntax: REGISTER <password>");
	return;
    }

    if ((n = FindNick(parv[1])) != NULL)
    {
	privmsg (u->nick, NickServ.nick, "Nickname already registered.");
	return;
    }

    AddNick (u, parv[1]);
    return;
}

static int
ns_drop (struct User *u, int parc, char *parv[])
{
    struct Nick *n;

    if ((parc < 2) && !(u->n))
    {
	privmsg(u->nick, NickServ.nick, "Syntax: DROP <password>");
	return;
    }

    if ((n = FindNick(u->nick)) == NULL)
    {
	privmsg(u->nick, NickServ.nick, "Nickname not registered");
	return;
    }

    if ((parc == 1) && (u->n))
    {
	DelNick (n->nick);
	privmsg (u->nick, NickServ.nick, "Nickname dropped.");
	u->n = NULL;
    }
    else if (parc == 2)
    {
#ifdef CRYPT_NICKPASS
	if (!strcmp(n->password, crypt(parv[2], n->password)))
#else
	if (!strcmp(n->password, parv[2]))
#endif
	{
	    DelNick (n->nick);
	    privmsg (u->nick, NickServ.nick, "Nickname dropped.");
	    u->n = NULL;
	    return;
	}
	else
	{
	    privmsg (u->nick, NickServ.nick, "Password Incorrect.");
	    return;
	}
    }

    /* NOT REACHED */
    return;
}

static int
ns_identify (struct User *u, int parc, char *parv[])
{
  struct NickKill *nickkill;
  dlink_node *dl, *tdl;
	struct Nick *nick;
	char	pass[MAXPASS];

	nick = FindNick(u->nick);

	if (parc < 2)
	{
		privmsg(u->nick, NickServ.nick, "Syntax: IDENTIFY <password>");
		return;
	}

	if (!nick)
	{
		privmsg(u->nick, NickServ.nick, RPL_NOT_REGISTERED, u->nick);
		return;
	}

	if (IsIdent(u))
	{
		privmsg(u->nick, NickServ.nick, RPL_ALREADY_IDENTIFIED);
		return;
	}

	strlcpy(pass, parv[1], MAXPASS+1);

#ifdef CRYPT_NICKPASS
	if (!strcmp(nick->password, crypt(pass, nick->password)))
#else
	if (!strcmp(nick->password, pass))
#endif
	{
		privmsg(u->nick, NickServ.nick, RPL_PASSWORD_ACCEPTED);
		nick->flags |= NS_IDENTIFIED;
		u->flags |= LVL_IDENT;
		nick->last_seen = time(NULL);
		strlcpy(nick->last_address, u->hostname, HOSTLEN+1);
		u->n = nick;

#ifdef IRCD_HYBSVS
		sts("SVSMODE %s :+R", u->nick);
#endif

#ifdef SPY
		Spy(NickServ.nick, "%s!%s@%s identified to nick %s", u->nick, u->username, u->hostname, u->n->nick);
#endif

		if (u->n->flags & NS_NOEXPIRE)
			notice(u, NickServ.nick, "No expiry is set");

		if (u->n->flags & NS_MEMOSIGNON)
			CheckMemos(u->n->nick);

    DLINK_FOREACH_SAFE(dl, tdl, nickkill_list.head)
    {
      nickkill = (struct NickKill *)dl->data;
      if (!strcasecmp(nickkill->nick, u->nick))
        dlink_delete(dl, &nickkill_list);
    }

		return;
	}
	else
	{
		privmsg(u->nick, NickServ.nick, "Password Incorrect");
		return;
	}

	/* NOT REACHED */
	return;
}

static int
ns_info (struct User *u, int parc, char *parv[])
{
  struct Nick *n;

  if (parc < 2)
  {
    notice(u, NickServ.nick, "Syntax: /msg %s \002INFO\002 <nickname>", NickServ.nick);
    return;
  }

  if ((n = FindNick(parv[1])) == NULL)
  {
    notice(u, NickServ.nick, RPL_NOT_REGISTERED, parv[1]);
    return;
  }

  notice(u, NickServ.nick, "          Nickname: %s %s", n->nick, (FindUser(n->nick))? "<< ONLINE >>" : "");
  notice(u, NickServ.nick, "        Registered: %s ago", timediff(time(NULL) - n->registered));
  notice(u, NickServ.nick, "         Last Seen: %s ago", timediff(time(NULL) - n->last_seen));
  notice(u, NickServ.nick, " Last Seen Address: %s", n->last_address);
  notice(u, NickServ.nick, "Last Seen Quit Msg: %s", n->last_quit);
  notice(u, NickServ.nick, "     Email Address: %s", (n->email)? n->email : "None Set");
  notice(u, NickServ.nick, "               URL: %s", (n->url)? n->url : "None Set");
  notice(u, NickServ.nick, "  Nickname Options: %s", NickOptions(n->flags));

  return;
}

char *NickOptions (int flags)
{
  static char optstring[BUFSIZE];
  optstring[0] = '\0';

  if (flags & NS_NOEXPIRE)
    if (optstring[0] = '\0')
      strcpy(optstring, "NoExpire");
    else
      strcat(optstring, ", NoExpire");

  if (flags & NS_AUTOMASK)
    if (optstring[0] == '\0')
      strcpy(optstring, "AutoMask");
    else
      strcat(optstring, ", AutoMask");

  if (flags & NS_ALLOWMEMO)
    if (optstring[0] == '\0')
      strcpy(optstring, "AllowMemo");
    else
      strcat(optstring, ", AllowMemo");

  if (flags & NS_KILL)
    if (optstring[0] == '\0')
      strcpy(optstring, "AutoKill");
    else
      strcat(optstring, ", AutoKill");

  if (flags & NS_SECURE)
    if (optstring[0] == '\0')
      strcpy(optstring, "Secure");
    else
      strcat(optstring, ", Secure");

  if (flags & NS_MEMONOTIFY)
    if (optstring[0] == '\0')
      strcpy(optstring, "MemoNotify");
    else
      strcat(optstring, ", MemoNotify");

  if (flags & NS_MEMOSIGNON)
    if (optstring[0] == '\0')
      strcpy(optstring, "MemoSignon");
    else
      strcat(optstring, ", MemoSignon");

  return optstring;
}

int
IsLinked (char *nick, char *target)
{
  dlink_node *dl;
  struct LinkedNicks *link;

  DLINK_FOREACH(dl, linked_nicks.head)
  {
    link = (struct LinkedNicks *)dl->data;

    if (!strcasecmp(link->master, nick))
      if (!strcasecmp(link->nick, target))
        return 1;
  }

  return 0;
}

/* AddNick - Add new nickname registration
**
*/
void AddNick (struct User *u, char *password)
{
    struct Nick *n;
    dlink_node *dl;

    time_t now = time(NULL);

#ifdef CRYPT_NICKPASS
    char * salt;
    salt = make_salt();
#endif

    if ((n = FindNick (u->nick)) != NULL)
    {
	privmsg(u->nick, NickServ.nick, "Nickname already registered.");
	return;
    }


    dl = dlink_create();
    n = calloc(1, sizeof(struct Nick));

    strlcpy(n->nick, u->nick, NICKLEN+1);
#ifdef CRYPT_NICKPASS
    strlcpy(n->password, crypt(password, salt), MAXPASS+1);
#else
    strlcpy(n->password, password, MAXPASS+1);
#endif
    strlcpy(n->last_address, u->hostname, HOSTLEN+1);

    n->flags |= NickServ.flags|NS_IDENTIFIED;
    n->last_seen = now;
    n->registered = now;
    n->expire = NickServ.expire;
	u->flags |= LVL_IDENT;

    dlink_add(n, dl, &nick_list);

    privmsg(u->nick, NickServ.nick, "Registered %s with password %s", n->nick, password);
#ifdef IRCD_HYBSVS
    sts("SVSMODE %s :+R", n->nick);
#endif

    u->n = n;
    return;
}

/* FindNick - Find existing nickname registration
**
*/
struct Nick *FindNick (char *nick)
{
    dlink_node *dl;
    struct Nick *n;

    DLINK_FOREACH(dl, nick_list.head)
    {
	n = (struct Nick *)dl->data;
	if (!strcasecmp(n->nick, nick))
	    return n;
    }

    debug("D: Nick %s not found.", nick);

    return 0;
}

/* DelNick - Remove nickname registration
**
*/
void DelNick (char *nick)
{
    dlink_node *dl;
    struct Nick *n;
  
    DLINK_FOREACH(dl, nick_list.head)
    {
	n = (struct Nick *)dl->data;
	if (!strcasecmp(nick, n->nick))
	{
	    debug("D: DelNick(%s): Successful", n->nick);
	    dlink_delete(dl, &nick_list);
	    dlink_free(dl);
	    free(n);
	    return;
	}
    }
    debug("D: DelNick(%s): Failed");
    return;
}
#endif /* NICKSERVICES */
